#include "uart.h"

#include "async.h"
#include "gpio.h"
#include "ring_buffer.h"
#include "trace.h"
#include "xuartlite.h"

namespace {
using async::AwaitableType;

constexpr uintptr_t kUart0BaseAddress = 0x40001000;
XUartLite uart0_inst;
XUartLite_Config uart0_config = {
    .DeviceId = 0,
    .RegBaseAddr = kUart0BaseAddress,
    .BaudRate = 115200,
    .UseParity = false,
    .DataBits = 8,
};

constexpr size_t kUartTxBufferSize = 256;
std::array<std::byte, kUartTxBufferSize> tx_buffer = {};
RingBuffer tx_ring_buffer{.buffer = tx_buffer};

XUartLite* uart0 = &uart0_inst;
}  // namespace

void InitUarts() {
    XUartLite_CfgInitialize(uart0, &uart0_config, uart0_config.RegBaseAddr);

    XUartLite_SetSendHandler(uart0, HandleUartTxFromIsr, nullptr);
    XUartLite_SetRecvHandler(uart0, HandleUartRxFromIsr, nullptr);
    XUartLite_EnableInterrupt(uart0);
}

// xuartlite private header stuff
extern "C" uint8_t XUartLite_GetSR(XUartLite* InstancePtr);
#define XUL_SR_TX_FIFO_EMPTY 0x04 /* transmit FIFO empty */

void UartWriteCrash(std::span<const std::byte> data) {
    while (data.size() > 0) {
        while ((XUartLite_GetSR(uart0) & XUL_SR_TX_FIFO_EMPTY) == 0) {
        }
        auto* dat =
            reinterpret_cast<uint8_t*>(const_cast<std::byte*>(data.data()));
        uint8_t sent = XUartLite_Send(uart0, dat, data.size());
        data = data.subspan(sent);
    }
    while ((XUartLite_GetSR(uart0) & XUL_SR_TX_FIFO_EMPTY) == 0) {
    }
}

async::task<> UartWrite(std::span<const std::byte> data) {
    while (!tx_ring_buffer.Store(data)) {
        co_await async::await(AwaitableType::kUartTx);
    }

    if (!XUartLite_IsSending(uart0)) {
        XUartLite_Send(uart0, tx_ring_buffer.RawReadPointer(),
                       tx_ring_buffer.ContiguousAvailableData());
    }
}

void UartReadBlocking(std::span<std::byte> data) {
    size_t bytes_received = 0;
    while (bytes_received < data.size()) {
        auto* buffer = reinterpret_cast<uint8_t*>(data.data() + bytes_received);
        bytes_received +=
            XUartLite_Recv(uart0, buffer, data.size() - bytes_received);
    }
}

void UartWriteBlocking(std::span<const std::byte> data) {
    while (!tx_ring_buffer.Store(data)) {
    }

    if (!XUartLite_IsSending(uart0)) {
        XUartLite_Send(uart0, tx_ring_buffer.RawReadPointer(),
                       tx_ring_buffer.ContiguousAvailableData());
    }
}

async::task<buffer> UartRead(int size) {
    auto buff = buffer::make(size);
    auto* data = reinterpret_cast<uint8_t*>(buff.data.data());
    size_t received = XUartLite_Recv(uart0, data, buff.data.size());
    if (received < buff.data.size()) {
        co_await async::await(AwaitableType::kUartRx);
    }
    co_return buff;
}

void HandleUartTxFromIsr(void*, unsigned int transmitted) {
    tracing::trace(tracing::TraceEvent::kUartTxCb);
    tx_ring_buffer.Pop(transmitted);
    if (tx_ring_buffer.AvailableData() > 0) {
        XUartLite_Send(uart0, tx_ring_buffer.RawReadPointer(),
                       tx_ring_buffer.ContiguousAvailableData());
    }
    async::resume(AwaitableType::kUartTx);
}

void HandleUartRxFromIsr(void*, unsigned int) {
    tracing::trace(tracing::TraceEvent::kUartRxCb);
    async::resume(AwaitableType::kUartRx);
}

void HandleUartIsr() { XUartLite_InterruptHandler(uart0); }
